package com.tencent.karaoketv.common.monitor

import android.content.Context
import android.content.Intent
import android.graphics.Color
import android.graphics.PixelFormat
import android.os.Build
import android.os.Handler
import android.os.Message
import android.provider.Settings
import android.view.Choreographer
import android.view.Gravity
import android.view.WindowManager
import android.widget.TextView
import com.tencent.karaoketv.common.KaraokeContext
import com.tencent.qqmusiccommon.util.MLog
import com.tencent.qqmusicsdk.player.mediaplayer.VideoPlayer
import com.tencent.karaoketv.common.sp.TvPreferences
import com.tencent.karaoketv.common.sp.TvPreferences.KEY_REPORT_FRAME_RATE

/**
 * Created by vienwang on 2018/2/2.
 * 帧率检测
 */

class FrameMonitor{

    val TAG = "FrameMonitor"

    val TYPE_FRAME_MONITOR = 0

    val TYPE_FRAME_REPORT = 1

    private var mFrameType:Int = 0

    /* 检测时间,纳秒为单位，实时监控 */
    private val MONITOR_TIME = 1000000000

    /* 检测时间,纳秒为单位，上报帧率 */
    private val FRAME_REPORT_TIME_NANO = 15000000000

    /* 检测时间,纳秒为单位，上报帧率 */
    private val FRAME_REPORT_TIME_SECOND = FRAME_REPORT_TIME_NANO/1000000000

    /* 检测次数 */
    private var count = 0

    /* 开始计算单位帧率（默认获取1s刷新UI的次数）的时间起点,纳秒为单位 */
    private var startFrameRefreshTimeNanos: Long = 0

    /* 当前的检测状态是否在停止状态 */
    private var isStopped = true

    private var mWindowManager: WindowManager? = null

    private var params: WindowManager.LayoutParams? = null

    /* 当前的帧率是否正在显示 */
    private var isShown: Boolean = false

    private var mView: TextView? = null

    /* MV丢帧的数目 */
    private var mMVDropFrameCount = 0

    /* 用来上报帧率时候使用的缓存数组 */
    private var mFrameArray: ArrayList<Int> = ArrayList()

    /* 上报的帧率检测中标识是否正在采集帧率 */
    private var isStartMonitorInTime:Boolean = false

    /* 上报的帧率均值 */
    private var mFrameAverage:Int = 0

    /* 帧率检测用到的检测间隔计数，每1s测试一次帧率 */
    private var mMonitorCount:Int = 0

    private var callBack:Choreographer.FrameCallback? = null

    private var mvDropFrameRate:Int = 0

    companion object {
        val sInstance by lazy { FrameMonitor() }
    }

    /**
     * 开启检测
     */
    fun start() {
        if(Build.VERSION.SDK_INT < Build.VERSION_CODES.JELLY_BEAN){
            return
        }
        checkCallBack()
        isStopped = false
        if(callBack != null) {
            Choreographer.getInstance().postFrameCallback(callBack)
        }
        showPopupWindow(KaraokeContext.getApplicationContext(), "Current Frame Rate: 0")
        MLog.d(TAG,"frame monitor start")
    }

    private fun checkCallBack(){
        if(Build.VERSION.SDK_INT < Build.VERSION_CODES.JELLY_BEAN){
            return
        }
        if(callBack == null){
            callBack = object : Choreographer.FrameCallback {
                override fun doFrame(frameTimeNanos: Long){
                    if (startFrameRefreshTimeNanos == 0L) {
                        startFrameRefreshTimeNanos = frameTimeNanos
                        if(Build.VERSION.SDK_INT >= Build.VERSION_CODES.JELLY_BEAN){
                            Choreographer.getInstance().postFrameCallback(this)
                        }
                        return
                    }
                    if(mFrameType == TYPE_FRAME_MONITOR) {
                        if ((frameTimeNanos - startFrameRefreshTimeNanos) < MONITOR_TIME) {
                            count++
                        } else {
                            setShowTxt("Current Frame Rate:$count\nMV dropped frames:${VideoPlayer.mDroppedFrameCount}")
                            count = 0
                            mMonitorCount = 0
                            startFrameRefreshTimeNanos = frameTimeNanos
                        }
                    }else if(mFrameType == TYPE_FRAME_REPORT){
                        if(isStartMonitorInTime){
                            if ((frameTimeNanos - startFrameRefreshTimeNanos) < MONITOR_TIME) {
                                count++
                            } else {
                                mFrameArray.add(count)
                                MLog.d(TAG,"mFrameArray added:" + count)
                                startFrameRefreshTimeNanos = frameTimeNanos
                                count= 0
                                if(mFrameArray.size >= 15){
                                    TvPreferences.getInstance().setLongValue(KEY_REPORT_FRAME_RATE,System.currentTimeMillis())
                                    sInstance.stop()
                                    isStartMonitorInTime = false
                                }
                            }
                        }
                    }
                    if (!isStopped) {
                        if(Build.VERSION.SDK_INT >= Build.VERSION_CODES.JELLY_BEAN){
                            Choreographer.getInstance().postFrameCallback(this)
                        }
                    }
                }
            }
        }
    }

    fun getInstance(): FrameMonitor {
        return sInstance
    }

    /**
     * 停止
     */
    fun stop() {
        if(Build.VERSION.SDK_INT < Build.VERSION_CODES.JELLY_BEAN || isStopped){
            return
        }
        mHandler.removeMessages(0)
        startFrameRefreshTimeNanos = 0
        isStopped = true
        count = 0
        hidePopupWindow()
        MLog.d(TAG,"frame monitor stop")
    }

    /**
     *
     * 计算跳过多少帧
     * @param start
     * @param end
     * @param devicefreshRate
     * @return
     */
    private fun skipFrameCount(start: Long, end: Long, devicefreshRate: Float): Int {
        var count = 0
        val diffNs = end - start
        val diffMs = Math.round(diffNs / 1000000.0f).toLong()
        val dev = Math.round(devicefreshRate).toLong()
        if (diffMs > dev) {
            val skipCount = diffMs / dev
            count = skipCount.toInt()
        }
        return count
    }

    /**
     * 更新MV掉帧
     * @param count 掉帧数目
     */
    public fun setMVDropFrame(count : Int){
        mMVDropFrameCount = count
    }

    /**
     * 显示弹出框
     *
     * @param context
     */
    private fun showPopupWindow(context: Context, showText: String) {
        if (isShown || checkPermission(context)) {
            return
        }

        isShown = true
        // 获取WindowManager
        mWindowManager = context
                .getSystemService(Context.WINDOW_SERVICE) as WindowManager
        mView = setUpView(context, showText)

        params = WindowManager.LayoutParams()
        // 类型，系统提示以及它总是出现在应用程序窗口之上。
        params?.let{
            it.type = WindowManager.LayoutParams.TYPE_SYSTEM_ALERT or WindowManager.LayoutParams.TYPE_SYSTEM_OVERLAY

            // 设置flag
            val flags = canTouchFlags

            // | WindowManager.LayoutParams.FLAG_NOT_FOCUSABLE;
            // 如果设置了WindowManager.LayoutParams.FLAG_NOT_FOCUSABLE，弹出的View收不到Back键的事件
            it.flags = flags
            // 不设置这个弹出框的透明遮罩显示为黑色
            it.format = PixelFormat.TRANSLUCENT
            // FLAG_NOT_TOUCH_MODAL不阻塞事件传递到后面的窗口
            // 设置 FLAG_NOT_FOCUSABLE 悬浮窗口较小时，后面的应用图标由不可长按变为可长按
            // 不设置这个flag的话，home页的划屏会有问题
            it.width = WindowManager.LayoutParams.WRAP_CONTENT
            it.height = WindowManager.LayoutParams.WRAP_CONTENT
            it.gravity = Gravity.LEFT
        }

        mWindowManager?.addView(mView, params)
    }

    private val canTouchFlags = WindowManager.LayoutParams.FLAG_NOT_FOCUSABLE or WindowManager.LayoutParams.FLAG_NOT_TOUCH_MODAL

    private val notTouchFlags = WindowManager.LayoutParams.FLAG_NOT_FOCUSABLE or WindowManager.LayoutParams.FLAG_NOT_TOUCHABLE

    private fun checkPermission(context: Context):Boolean{
        if(Build.VERSION.SDK_INT >= Build.VERSION_CODES.M){
            if (!Settings.canDrawOverlays(context)) {
                val intent = Intent(Settings.ACTION_MANAGE_OVERLAY_PERMISSION)
                intent.addFlags(Intent.FLAG_ACTIVITY_NEW_TASK)
                context.startActivity(intent)
                return true
            }
        }
        return false
    }

    /**
     *  用来处理上报的帧率测试异步执行
     *  */
    var mHandler: Handler = object : Handler() {

        override fun handleMessage(msg: Message) {
            super.handleMessage(msg)
            MLog.d(TAG,"handleMessage start send message post~")
            startFrameRefreshTimeNanos = 0
            isStartMonitorInTime = true
            if(Build.VERSION.SDK_INT >= Build.VERSION_CODES.JELLY_BEAN && callBack != null){
                Choreographer.getInstance().postFrameCallback(callBack)
            }

        }
    }

    /**
     * 设置是否可响应点击事件
     *
     * @param isTouchable
     */
    fun setTouchable(isTouchable: Boolean) {
        if (isTouchable) {
            params?.let { it.flags = canTouchFlags }
        } else {
            params?.let { it.flags = notTouchFlags }
        }
        mWindowManager?.updateViewLayout(mView, params)

    }

    fun switchFrameMonitorForRealTime(type:Int) {
        //若是上报测试帧率则一天只测试一次
        mFrameType = type
        if (isShown) {
            stop()
        } else {
            start()
        }
    }

    /**
     * 开始检测帧率，规则如下：
     * 在有ＭＶ和歌词的情况下启动检测，并且检测间隔必须超过一天
     */
    fun detectFrameForReport():Boolean {
        if (isShown) {
            //如果当前正在进行帧率测试，则返回
            MLog.d(TAG,"frame monitor is on!!!")
            return false
        }
        mFrameType = TYPE_FRAME_REPORT
        if(System.currentTimeMillis() - TvPreferences.getInstance().getLongValue(KEY_REPORT_FRAME_RATE) < 86400000){
            return false
        }
        isStopped = false
        checkCallBack()
        sInstance.mHandler.sendEmptyMessageDelayed(0,30000)
        return true
    }

    /**
     * 隐藏弹出框
     */
    private fun hidePopupWindow() {
        if (isShown && null != mView) {
            mWindowManager?.removeView(mView)
            isShown = false
        }
    }

    private fun setShowTxt(txt: String) {
        mView?.text = txt
    }

    private fun setUpView(context: Context, showtxt: String): TextView {
        val showTv = TextView(context)
        showTv.text = showtxt
        showTv.setTextColor(Color.RED)
        return showTv
    }

    /**
     * 当前是否在展示实时帧率
     */
    fun isShow():Boolean{
        return isShown
    }

    fun getFrameRateStr():String{

        var frameStr = ""

        for(i in mFrameArray){
            frameStr += i
            frameStr += "_"
        }
        MLog.d(TAG,"frameStr:" + frameStr)
        return frameStr
    }

    fun resetFrameRateStr(){
        mFrameArray.clear()
    }

    fun getMVFrameRate():Int{
        MLog.d(TAG,"getMVFrameRate:" + mvDropFrameRate)
        return mvDropFrameRate
    }

    fun setMVFrameRate(farme:Int ){
        MLog.d(TAG,"setMVFrameRate:" + farme)
        mvDropFrameRate = farme
    }
}



